/* packet-HSMS.c
 * Routines for HSMS protocol packet disassembly
 * By Dirk Voss <DirkVoss@DirkVoss.com>
 * Copyright 2012 Dirk Voss
 *
 * $Id: packet-HSMS.c 42367 2012-04-30 23:03:51Z guy $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/dissectors/packet-tcp.h>

#define FRAME_HEADER_LEN	4
#define HSMS_PORT 5000
#define offset_PType	8
#define offset_SType	9
#define HSMS_WBIT_FLAG	0x80


static gint ett_HSMS = -1;
static int	proto_HSMS = -1;

static int	hf_HSMS_pdu_function = -1;
static int	hf_HSMS_pdu_header_byte_2 = -1;
static int	hf_HSMS_pdu_header_byte_3 = -1;
static int	hf_HSMS_pdu_length = -1;
static int	hf_HSMS_pdu_ptype = -1;
static int	hf_HSMS_pdu_session_id = -1;
static int	hf_HSMS_pdu_stream = -1;
static int	hf_HSMS_pdu_stype = -1;
static int	hf_HSMS_pdu_system_bytes = -1;
static int	hf_HSMS_pdu_wbit = -1;

static guint	HSMS_tcp_port	= HSMS_PORT;							/* overwritten by preferences */
static gboolean HSMS_desegment	= TRUE;									/* overwritten by preferences */

static const value_string controlmessagenames[] = {
	{ 1, "Select Request" },
	{ 2, "Select Response" },
	{ 3, "Deselect Request" },
	{ 4, "Deselect Response" },
	{ 5, "Linktest Request" },
	{ 6, "Linktest Response" },
	{ 7, "Reject Request" },
	{ 8, "(not used)" },
	{ 9, "Separate Request" },
	{10, "(not used)" },
	{ 0, NULL }
};

/* Forward declaration(s) needed */
void proto_reg_handoff_HSMS(void);


/* Dissect fully reassembled HSMS messages */
static void
dissect_HSMS_message(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	gint offset = 0;
	guint8 PType;
	guint8 SType;


	PType = tvb_get_guint8(tvb, offset_PType);
	SType = tvb_get_guint8(tvb, offset_SType);


	col_set_str(pinfo->cinfo, COL_PROTOCOL, "HSMS");
	col_clear(pinfo->cinfo, COL_INFO);					/* clear out info column */

	if (tree) { /* we are being asked for details */
		proto_item *ti = NULL;
		proto_tree *HSMS_tree = NULL;

		/* Add the HSMS protocol to the display tree */
		ti = proto_tree_add_item(tree, proto_HSMS, tvb, 0, -1, ENC_NA);
		HSMS_tree = proto_item_add_subtree(ti, ett_HSMS);

		/* Display the Message Length */
		proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_length,			tvb, offset, 4, ENC_BIG_ENDIAN);
		offset += 4;

		/* Display the Message Header */
		proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_session_id,		tvb, offset, 2, ENC_BIG_ENDIAN);
		offset += 2;
		
		if (PType == 0) {							/* determine the type of message -- SECS-II, HSMS Control, or Unknown */
			if (SType == 0) {
				/* SECS-II Message */
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_wbit,		tvb, offset, 1, ENC_BIG_ENDIAN);
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_stream,		tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_function,	tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				offset += 2;		/* skip the PType and SType bytes */
				/* Display the Message Text -- by passing the "buck" to the SECS-II protocol */
			}
			else {
				/* HSMS Control Message */
						/* could label as "PType or SType of message being rejected if SType = 7 */
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_header_byte_2,	tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
						/* could label as "Select Status" (SType = 2), "Deselect Status" (4), "Reason Code (7) instead of Header Byte 3 */
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_header_byte_3,	tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_ptype,			tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_stype,			tvb, offset, 1, ENC_BIG_ENDIAN);	/* range check needed? */
				offset += 1;
			}
		}
		else {
			/* Unknown message type */
		}

		proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_system_bytes,			tvb, offset, 4, ENC_BIG_ENDIAN);	
	}
}


/* Determine PDU length of protocol HSMS */
/*		how does it how many bytes the length field is? */
static guint 
get_HSMS_message_len(packet_info *pinfo, tvbuff_t *tvb, int offset)
{
	return (	(guint)tvb_get_ntohl(tvb, offset) + FRAME_HEADER_LEN);						/* length is at offset 0 */
}


/* reassemble HSMS messages */
static void
dissect_HSMS(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{	
	tcp_dissect_pdus(	tvb, 
						pinfo, 
						tree, 
						HSMS_desegment,					/* by default TRUE, but allow the user override */
						FRAME_HEADER_LEN,
						get_HSMS_message_len,
						dissect_HSMS_message);
}


void
proto_register_HSMS(void)
{
	static hf_register_info hf[] = {
		{ &hf_HSMS_pdu_function,
			{ "HSMS PDU SECS-II Function", "HSMS.function",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_header_byte_2,
			{ "HSMS PDU Header Byte 2", "HSMS.header_byte_2",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_header_byte_3,
			{ "HSMS PDU Header Byte 3", "HSMS.header_byte_3",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_length,
			{ "HSMS PDU Message Length", "HSMS.length",
			FT_UINT32, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_ptype,
			{ "HSMS PDU PType", "HSMS.ptype",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_session_id,
			{ "HSMS PDU Session ID", "HSMS.session_id",
			FT_UINT16, BASE_HEX,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_stream,
			{ "HSMS PDU SECS-II Stream", "HSMS.stream",
			FT_UINT8, BASE_DEC,
			NULL, 0x7F,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_stype,
			{ "HSMS PDU SType", "HSMS.stype",
			FT_UINT8, BASE_DEC,
			VALS(controlmessagenames), 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_system_bytes,
			{ "HSMS PDU System Bytes", "HSMS.system_bytes",
			FT_UINT32, BASE_HEX,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_wbit,
			{ "HSMS PDU WBit", "HSMS.wbit",
			FT_BOOLEAN, 8,
			NULL, HSMS_WBIT_FLAG,
			NULL, HFILL }
		}
	};

	/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_HSMS
	};

	module_t *HSMS_module;


    proto_HSMS = proto_register_protocol(
		"HSMS Protocol",
		"HSMS",
		"hsms");

	proto_register_field_array(proto_HSMS, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	/* HSMS Protocol Preferences */
	HSMS_module = prefs_register_protocol(proto_HSMS, proto_reg_handoff_HSMS);

	prefs_register_uint_preference(HSMS_module, 
		"tcp_port",									/* name -- preference file field name */
		"HSMS TCP Port",							/* title -- Field title */
		NULL,										/* description -- tooltip */
		10,											/* base */
		&HSMS_tcp_port);							/* variable */
    prefs_register_bool_preference(HSMS_module, 
		"reassemble",
        "Reassemble all HSMS messages spanning multiple TCP packets",
        "Dont assume 1 HSMS message per TCP packet or 1 TCP packet per HSMS message",
        &HSMS_desegment);

}

/* This function is registered as a prefs callback. It is also called by preferences whenever "Apply" is pressed; it should accommodate 
   being called more than once.
*/
void
proto_reg_handoff_HSMS(void)
{
    static dissector_handle_t	HSMS_handle;
	static gboolean				initialized = FALSE;
    static gint					currentPort;

	if (!initialized) {
		HSMS_handle = create_dissector_handle(dissect_HSMS, proto_HSMS);
		initialized = TRUE;
		} 
	else {
 		/*
		  Performing registration functions which are dependent upon
		  prefs require de-registering everything which was associated
		  with the previous settings and re-registering using the new prefs
		  settings here. In general this means you need to keep track of
		  the PROTOABBREV_handle and the value the preference had at the time
		  you registered.  The PROTOABBREV_handle value and the value of the
		  preference can be saved using local statics in this function (proto_reg_handoff).
		*/
		dissector_delete_uint("tcp.port", currentPort, HSMS_handle);
	}
	currentPort = HSMS_tcp_port;
    dissector_add_uint("tcp.port", currentPort, HSMS_handle);
}
