/* packet-HSMS.c
 * Routines for HSMS protocol packet disassembly
 * By Dirk Voss <DirkVoss@DirkVoss.com>
 * Copyright 2012 Dirk Voss
 *
 * $Id: packet-HSMS.c 42367 2012-04-30 23:03:51Z guy $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/dissectors/packet-tcp.h>

#define FRAME_HEADER_LEN		4
#define HSMS_PORT				5000
#define	MAX_LINE_WIDTH			120
#define offset_Header_Byte_2	6
#define offset_Header_Byte_3	7
#define offset_Message_Length	0
#define offset_Session_ID		4
#define offset_PType			8
#define offset_SType			9
#define offset_System_Bytes		10
#define HSMS_WBIT_FLAG			0x80
#define MAXIMUM_GINT			0x7FFF

static gint ett_HSMS = -1;
static int	proto_HSMS = -1;

static int	hf_HSMS_pdu_function = -1;
static int	hf_HSMS_pdu_header_byte_2 = -1;
static int	hf_HSMS_pdu_header_byte_3 = -1;
static int	hf_HSMS_pdu_length = -1;
static int	hf_HSMS_pdu_ptype = -1;
static int	hf_HSMS_pdu_session_id = -1;
static int	hf_HSMS_pdu_stream = -1;
static int	hf_HSMS_pdu_stype = -1;
static int	hf_HSMS_pdu_system_bytes = -1;
static int	hf_HSMS_pdu_wbit = -1;
static int	hf_HSMS_pdu_SECS_message = -1;

static guint	HSMS_tcp_port		= HSMS_PORT;							/* overwritten by preferences */
static gboolean HSMS_desegment		= TRUE;									/* overwritten by preferences */
static gboolean HSMS_debug_output	= FALSE;

static const value_string controlmessagenames[] = {
	{ 1, "Select Request" },
	{ 2, "Select Response" },
	{ 3, "Deselect Request" },
	{ 4, "Deselect Response" },
	{ 5, "Linktest Request" },
	{ 6, "Linktest Response" },
	{ 7, "Reject Request" },
	{ 8, "(not used)" },
	{ 9, "Separate Request" },
	{10, "(not used)" },
	{ 0, NULL }
};

static char testing[] = "testing";




/* Forward declaration(s) needed */
void proto_reg_handoff_HSMS(void);
void display_SECSmsg(guint8 *SECSmsg, gint length);

/* Dissect fully reassembled HSMS messages */
static void
dissect_HSMS_message(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	gint		offset = 0;
	guint8		Header_Byte_2;
	gboolean	WBit;
	guint8		Stream;
	guint8		Header_Byte_3;
	guint32		Message_Length;
	guint8		PType;
	guint16		Session_ID;
	guint8		SType;
	guint32		System_Bytes;

	guint8		*SECSmsg;
	gint		Message_Length_gint;

	proto_item	*ti = NULL;
	proto_tree	*HSMS_tree = NULL;

	/*	Parse the HSMS message header for the overview pane display */
	col_set_str(pinfo->cinfo, COL_PROTOCOL, "HSMS");	/* mark this packet as "ours" */
	/* col_clear(pinfo->cinfo, COL_LENGTH);	*/
	col_clear(pinfo->cinfo, COL_INFO);					/* clear out info column */

	/* Add the HSMS protocol to the display tree */
	ti = proto_tree_add_item(tree, proto_HSMS, tvb, 0, -1, ENC_NA);
	HSMS_tree = proto_item_add_subtree(ti, ett_HSMS);


	/* Extract temporary field variables [ ]since we are using get_guintX calls, might be clearer to increment OFFSET */
	Message_Length	= tvb_get_ntohl(tvb, offset_Message_Length);
	Session_ID		= tvb_get_ntohs(tvb, offset_Session_ID);
	Header_Byte_2	= tvb_get_guint8(tvb, offset_Header_Byte_2);
	Stream			= Header_Byte_2 & 0x7F;
	WBit			= (Stream != Header_Byte_2);
	Header_Byte_3	= tvb_get_guint8(tvb, offset_Header_Byte_3);
	PType			= tvb_get_guint8(tvb, offset_PType);
	SType			= tvb_get_guint8(tvb, offset_SType);
	System_Bytes	= tvb_get_ntohl(tvb, offset_System_Bytes);

	if (PType == 0) 
		{							/* determine the type of message -- SECS-II, HSMS Control, or Unknown */
		if (SType == 0) 
			{
			/* SECS-II Message */
			col_set_str(pinfo->cinfo, COL_PROTOCOL, "SECS-II");
			/*	col_set_str(pinfo->cinfo, COL_LENGTH, Message_Length-10);			/* length of SECS-II message, not HSMS */
			col_add_fstr(pinfo->cinfo, COL_INFO, "S%u,F%u", Stream, Header_Byte_3);
			if (WBit) col_append_str(pinfo->cinfo, COL_INFO, " WBit"); 
			}
		else 
			{
			/* HSMS Control Message */
			/*	col_set_str(pinfo->cinfo, COL_LENGTH, Message_Length);	*/ 
			col_add_fstr(pinfo->cinfo, COL_INFO, "%s", val_to_str(SType, controlmessagenames, "Illegal Control Function %d"));
			if (SType == 2)
				col_append_fstr(pinfo->cinfo, COL_INFO, ", Status = %u", Header_Byte_3);		/* Select */
			else if (SType == 4)
				col_append_fstr(pinfo->cinfo, COL_INFO, ", Status = %u", Header_Byte_3);		/* Deselect */
			else if (SType == 7)
				col_append_fstr(pinfo->cinfo, COL_INFO, ", Reason = %u", Header_Byte_3);		/* Reject */
			}
		}
	else 
		{
		col_set_str(pinfo->cinfo, COL_INFO, "HSMS Protocol Error (PType != 0)");
		};



	if (tree) {													/* we are being asked for details */
		/* Display the Message Length */
		proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_length,			tvb, offset, FRAME_HEADER_LEN, ENC_BIG_ENDIAN);
		offset += FRAME_HEADER_LEN;

		/* Display the Message Header */
		proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_session_id,		tvb, offset, 2, ENC_BIG_ENDIAN);
		offset += 2;

		if (PType == 0) {							/* determine the type of message -- SECS-II, HSMS Control, or Unknown */
			if (SType == 0) {
				/* SECS-II Message */
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_wbit,		tvb, offset, 1, ENC_BIG_ENDIAN);
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_stream,		tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_function,	tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				offset += 2;		/* skip the PType and SType bytes */
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_system_bytes,			tvb, offset, 4, ENC_BIG_ENDIAN);	
				offset += 4;



				/* Display the Message Text -- by passing the "buck" to the SECS-II protocol */
				if (Message_Length < MAXIMUM_GINT)					/* [ ]take this check out */
					Message_Length_gint = Message_Length;
				else
					Message_Length = MAXIMUM_GINT;					/* may generate a false error, but an error is okay */
				SECSmsg = tvb_get_ephemeral_string(tvb, offset_Session_ID, Message_Length_gint);
				display_SECSmsg(SECSmsg, Message_Length_gint);

				if (0) /* HSMS_debug_output */ {
					proto_tree_add_string(HSMS_tree, hf_HSMS_pdu_SECS_message,	tvb, offset, 4, testing);
					proto_tree_add_text(HSMS_tree,								tvb, offset, 4, "Testing: %s", testing); 
				}


	/* IF statements for debugging/isolation purposes; "production" version will not an IF statement	*/
	/*			if ((Stream == 1) && (Header_Byte_3 == 13))			*/
	/*			if (Stream == 6)									*/
				
{ /* gimme a new block to declare variables */
#define MAX_LISTS_TRACKED	10
#define	MAX_FORMAT_WIDTH	10

gint	i;
gint	item_count[MAX_LISTS_TRACKED];
gint	indentation_spacing;
char	indentation_padding[] = "                                        ";
gint	list_count = 0;

gint	SECS_item_unused_chars;
gint	SECS_item_length;
gint	SECS_item_offset;
gint	number_of_length_bytes;
char	SECS_item[MAX_LINE_WIDTH];

char	formatting_buffer[MAX_LINE_WIDTH];
guint8	format_byte;
guint8	*format_ASCII;
gdouble	format_float;
guint64	format_int;


				for (i=0; i<MAX_LISTS_TRACKED; i++) {
					item_count[i] = -99;					/* 99 not significant; just want values < 0		*/
				}
				list_count = 0;
				SECS_item_offset = offset;					/* keep offset = start of SECS-II message "just in case" */
				while (SECS_item_offset < (Message_Length + FRAME_HEADER_LEN)) {
					if (HSMS_debug_output)
						proto_tree_add_text(HSMS_tree, tvb, offset, 4, "Item Offset %d", SECS_item_offset); 

					SECS_item_unused_chars = MAX_LINE_WIDTH;
					indentation_spacing = list_count << 2;								/* [  ]support 1-3 (2, 4, 8 spaces) */
					memset(SECS_item, '\0', MAX_LINE_WIDTH);
					strncpy(SECS_item, indentation_padding, indentation_spacing);	

					/* 1st byte of Item and List Header */
					format_byte	= tvb_get_guint8(tvb, SECS_item_offset);
					SECS_item_offset += 1;
					number_of_length_bytes = format_byte & 0x3;
					format_byte = format_byte >> 2;
					if (HSMS_debug_output)
						proto_tree_add_text(HSMS_tree, tvb, offset, 4, "Item Format %o (octal)", format_byte); 

					/* 2nd-4th bytes of Item and List Header */
					if (number_of_length_bytes == 1) {
						SECS_item_length = tvb_get_guint8(tvb, SECS_item_offset);
						SECS_item_offset += 1;
					}

					else if (number_of_length_bytes == 2) {
						SECS_item_length = tvb_get_ntohs(tvb, SECS_item_offset);
						SECS_item_offset += 2;
					}
					else if (number_of_length_bytes == 3) {
						SECS_item_length = tvb_get_ntoh24(tvb, SECS_item_offset);
						SECS_item_offset += 3;
					}
					else { /* protocol error of some sort, return to caller */
						/* [ ]update COL text to indicate punting */
						return;
					}

					/* SECS-II Item */
					if (format_byte == 0) { 						/* List item */
						list_count += 1;
						if (list_count >= MAX_LISTS_TRACKED) {
						/* message may be okay, but we're not */
							return;
						};	
						item_count[list_count] = SECS_item_length;
						g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<L,%d        ", SECS_item_length);
						strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);			/* not really checking for overflow here */

						/* deal with special case of L,0	*/
						if (SECS_item_length == 0) {
							list_count -= 1;
							item_count[list_count] -= 1;
							strncat(SECS_item, ">", MAX_LINE_WIDTH);
						}
					}

					else {	/* non-list item */
						switch (format_byte) {
/* [  ]Binary, ASCII -- need to worry about exceeding max line width		*/
						

							case 010:	/* Binary */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<B[%d]        ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);		/* not really checking for overflow here */
								i = 0;
								while (i < SECS_item_length) {
									format_byte = tvb_get_guint8(tvb, SECS_item_offset+i);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%02X ", format_byte);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
									i += 1;
								}
								strncat(SECS_item, ">", MAX_LINE_WIDTH);			/* need to special case B[0]	*/
								break;						
						

							case 011:	/* Boolean */
								strncat(SECS_item, "<BOOLEAN ", MAX_LINE_WIDTH);
								if (SECS_item_length == 0) {
									strncat(SECS_item, "null>", MAX_LINE_WIDTH);
								}
								else if (SECS_item_length == 1) {
									format_byte = tvb_get_guint8(tvb, SECS_item_offset);
									if (format_byte)
										strncat(SECS_item, "TRUE>", MAX_LINE_WIDTH);
									else strncat(SECS_item, "FALSE>", MAX_LINE_WIDTH);
								}
								else {
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "Unexpected/Unsupported item length of %d>", SECS_item_length);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								break;


							case 020:	/* ASCII */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<A[%d]        ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);		/* not really checking for overflow here */
								format_ASCII = tvb_get_ephemeral_string(tvb, SECS_item_offset, SECS_item_length);
								strncat(SECS_item, format_ASCII, MAX_LINE_WIDTH);
								strncat(SECS_item, """>", MAX_LINE_WIDTH);
								break;


							case 021:	/* JIS-8 */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<JIS[%d]    >", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								break;


							case 022:	/* 2-Byte Character */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<CHAR[%d]     ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);	
								i = 0;
								while (i < SECS_item_length) {
									format_int = tvb_get_ntohs(tvb, SECS_item_offset+i);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%04X ", format_int);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
									i += 2;
								}
								strncat(SECS_item, ">", MAX_LINE_WIDTH);			/* need to special case B[0]	*/
								break;						


							case 040:	/* 8-Byte Floating Point */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<F8[%d]       ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);		
								if (SECS_item_length == 8) {
									format_float = tvb_get_ntohieee_double(tvb, SECS_item_offset);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%f", format_float);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								else if (SECS_item_length == 0) {
									/* nuthin' to do but let the item be closed */
								}
								else { 
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "Unexpected/Unsupported item length of %d", SECS_item_length);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								strncat(SECS_item, """>", MAX_LINE_WIDTH);
								break;


							case 044:	/* 4-Byte Floating Point */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<F4[%d]       ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);		
								if (SECS_item_length == 4) {
									format_float = tvb_get_ntohieee_float(tvb, SECS_item_offset);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%f", format_float);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								else if (SECS_item_length == 0) {
									/* nuthin' to do but let the item be closed */
								}
								else { 
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "Unexpected/Unsupported item length of %d", SECS_item_length);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								strncat(SECS_item, """>", MAX_LINE_WIDTH);
								break;


							case 051:	/* 1-Byte Integer (unsigned) */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<U1[%d]       ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);		
								if (SECS_item_length == 1) {
									format_int = tvb_get_guint8(tvb, SECS_item_offset);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%d", format_int);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								else if (SECS_item_length == 0) {
									/* nuthin' to do but let the item be closed */
								}
								else { 
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "Unexpected/Unsupported item length of %d", SECS_item_length);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								strncat(SECS_item, """>", MAX_LINE_WIDTH);
								break;


							case 052:	/* 2-Byte Integer (unsigned) */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<U2[%d]       ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);		
								if (SECS_item_length == 2) {
									format_int = tvb_get_ntohs(tvb, SECS_item_offset);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%d", format_int);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								else if (SECS_item_length == 0) {
									/* nuthin' to do but let the item be closed */
								}
								else { 
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "Unexpected/Unsupported item length of %d", SECS_item_length);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								strncat(SECS_item, """>", MAX_LINE_WIDTH);
								break;


							case 054:	/* 4-Byte Integer (unsigned) */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<U4[%d]       ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);	
								if (SECS_item_length == 4) {
									format_int = tvb_get_ntohl(tvb, SECS_item_offset);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%d", format_int);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								else if (SECS_item_length == 0) {
									/* nuthin' to do but let the item be closed */
								}
								else { 
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "Unexpected/Unsupported item length of %d", SECS_item_length);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								strncat(SECS_item, """>", MAX_LINE_WIDTH);
								break;


							case 050:	/* 8-Byte Integer (unsigned) */
								g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "<U8[%d]       ", SECS_item_length);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);	
								if (SECS_item_length == 8) {
									format_int = tvb_get_ntoh64(tvb, SECS_item_offset);
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "%d", format_int);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								else if (SECS_item_length == 0) {
									/* nuthin' to do but let the item be closed */
								}
								else { 
									g_snprintf(formatting_buffer, MAX_FORMAT_WIDTH, "Unexpected/Unsupported item length of %d", SECS_item_length);
									strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								}
								strncat(SECS_item, """>", MAX_LINE_WIDTH);
								break;

							case 030:	/* 8-Byte Integer (signed) */

							case 031:	/* 1-Byte Integer (signed) */

							case 032:	/* 2-Byte Integer (signed) */

							case 034:	/* 4-Byte Integer (signed) */


							default:
								g_snprintf(formatting_buffer, MAX_LINE_WIDTH, "Unsupported Format Code %d", format_byte);
								strncat(SECS_item, formatting_buffer, MAX_LINE_WIDTH);
								break;

						} /* switch */

						item_count[list_count] -= 1;
						SECS_item_offset += SECS_item_length;
					} /* if-then-else */

					number_of_length_bytes += 1;			/* temporarily overload meaning; "point" to format & length bytes */
					proto_tree_add_string(HSMS_tree, hf_HSMS_pdu_SECS_message,	tvb, offset, number_of_length_bytes, SECS_item);


					/* if we have seen all the items in the list, then we need to "close" the list */
					/* do this last as this loop will terminate once all the items have been processed */
					while (item_count[list_count] == 0) {
						list_count -= 1;
						indentation_spacing = list_count << 2;								/* [  ]support 1-3 (2, 4, 8 spaces) */
						memset(SECS_item, '\0', MAX_LINE_WIDTH);
						strncpy(SECS_item, indentation_padding, indentation_spacing);
						strncat(SECS_item, ">", SECS_item_unused_chars);
						item_count[list_count] -= 1;
						proto_tree_add_string(HSMS_tree, hf_HSMS_pdu_SECS_message,	tvb, offset, SECS_item_length, SECS_item);
					}

				} /* while (SECS_item_offset < (Message_Length + FRAME_HEADER_LEN)) */



} /* SECS-II Message display block */





			}
			else {
				/* HSMS Control Message */
						/* could label as "PType or SType of message being rejected if SType = 7 */
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_header_byte_2,	tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
						/* could label as "Select Status" (SType = 2), "Deselect Status" (4), "Reason Code (7) instead of Header Byte 3 */
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_header_byte_3,	tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_ptype,			tvb, offset, 1, ENC_BIG_ENDIAN);
				offset += 1;
				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_stype,			tvb, offset, 1, ENC_BIG_ENDIAN);	/* range check needed? */
				offset += 1;

				proto_tree_add_item(HSMS_tree, hf_HSMS_pdu_system_bytes,	tvb, offset, 4, ENC_BIG_ENDIAN);	
			}
		}
		else {
			/* Unknown message type */
		};

		
		
	}	/* if (tree) */
}


static void
display_SECSmsg(guint8 *SECSmsg, gint length)
{
	return;
}

/* Determine PDU length of protocol HSMS */
/*		how does it how many bytes the length field is? */
static guint 
get_HSMS_message_len(packet_info *pinfo, tvbuff_t *tvb, int offset)
{
	return (	(guint)tvb_get_ntohl(tvb, offset) + FRAME_HEADER_LEN);						/* length is at offset 0 */
}


/* reassemble HSMS messages */
static void
dissect_HSMS(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{	
	tcp_dissect_pdus(	tvb, 
						pinfo, 
						tree, 
						HSMS_desegment,					/* by default TRUE, but allow the user override */
						FRAME_HEADER_LEN,
						get_HSMS_message_len,
						dissect_HSMS_message);
}


void
proto_register_HSMS(void)
{
	static hf_register_info hf[] = {
		{ &hf_HSMS_pdu_function,
			{ "HSMS PDU SECS-II Function", "HSMS.function",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_header_byte_2,
			{ "HSMS PDU Header Byte 2", "HSMS.header_byte_2",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_header_byte_3,
			{ "HSMS PDU Header Byte 3", "HSMS.header_byte_3",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_length,
			{ "HSMS PDU Message Length", "HSMS.length",
			FT_UINT32, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_ptype,
			{ "HSMS PDU PType", "HSMS.ptype",
			FT_UINT8, BASE_DEC,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_session_id,
			{ "HSMS PDU Session ID", "HSMS.session_id",
			FT_UINT16, BASE_HEX,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_stream,
			{ "HSMS PDU SECS-II Stream", "HSMS.stream",
			FT_UINT8, BASE_DEC,
			NULL, 0x7F,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_stype,
			{ "HSMS PDU SType", "HSMS.stype",
			FT_UINT8, BASE_DEC,
			VALS(controlmessagenames), 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_system_bytes,
			{ "HSMS PDU System Bytes", "HSMS.system_bytes",
			FT_UINT32, BASE_HEX,
			NULL, 0x0,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_wbit,
			{ "HSMS PDU WBit", "HSMS.wbit",
			FT_BOOLEAN, 8,
			NULL, HSMS_WBIT_FLAG,
			NULL, HFILL }
		},
		{ &hf_HSMS_pdu_SECS_message,
			{ " ", "HSMS.SECS_item",
			FT_STRING, BASE_NONE,
			NULL, 0x0,
			NULL, HFILL }
		}
	};

	/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_HSMS
	};

	module_t *HSMS_module;


    proto_HSMS = proto_register_protocol(
		"HSMS Protocol",
		"HSMS",
		"hsms");

	proto_register_field_array(proto_HSMS, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	/* HSMS Protocol Preferences */
	HSMS_module = prefs_register_protocol(proto_HSMS, proto_reg_handoff_HSMS);

	prefs_register_uint_preference(HSMS_module, 
		"tcp_port",									/* name -- preference file field name */
		"HSMS TCP Port",							/* title -- Field title */
		NULL,										/* description -- tooltip */
		10,											/* base */
		&HSMS_tcp_port);							/* variable */
    prefs_register_bool_preference(HSMS_module, 
		"reassemble",
        "Reassemble all HSMS messages spanning multiple TCP packets",
        "Dont assume 1 HSMS message per TCP packet or 1 TCP packet per HSMS message",
        &HSMS_desegment);

}

/* This function is registered as a prefs callback. It is also called by preferences whenever "Apply" is pressed; it should accommodate 
   being called more than once.
*/
void
proto_reg_handoff_HSMS(void)
{
    static dissector_handle_t	HSMS_handle;
	static gboolean				initialized = FALSE;
    static gint					currentPort;

	if (!initialized) {
		HSMS_handle = create_dissector_handle(dissect_HSMS, proto_HSMS);
		initialized = TRUE;
		} 
	else {
 		/*
		  Performing registration functions which are dependent upon
		  prefs require de-registering everything which was associated
		  with the previous settings and re-registering using the new prefs
		  settings here. In general this means you need to keep track of
		  the PROTOABBREV_handle and the value the preference had at the time
		  you registered.  The PROTOABBREV_handle value and the value of the
		  preference can be saved using local statics in this function (proto_reg_handoff).
		*/
		dissector_delete_uint("tcp.port", currentPort, HSMS_handle);
	}
	currentPort = HSMS_tcp_port;
    dissector_add_uint("tcp.port", currentPort, HSMS_handle);
}
